/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2000
 *
 * Copyright (c) 1999 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/objects.h>
#include <openssl/pem.h>
#include <openssl/asn1.h>
#include <openssl/asn1_mac.h>
#include <openssl/txt_db.h>

#include <ocsp.h> 
#include <ocsp_data.h>

/* Static Data Needed */
#define OCSP_REQUEST_sign(o,pkey,md) ASN1_sign((int(*)())i2d_OCSP_REQINFO, o->optionalSignature->signatureAlgorithm,NULL, o->optionalSignature->signature,(char *)o->tbsRequest,pkey,md)

/*
** Functions: main and tool func to manipulate the DB file
** =======================================================
*/

static void lookup_fail(char *name, char *tag)
        {
        BIO_printf(bio_err,"variable lookup failed for %s::%s\n",name,tag);
        }

static unsigned long index_serial_hash(char **a)
        {
        char *n;

        n=a[DB_serial];
        while (*n == '0') n++;
        return(lh_strhash(n));
        }

static int index_serial_cmp(char **a, char **b)
        {
        char *aa,*bb;

        for (aa=a[DB_serial]; *aa == '0'; aa++);
        for (bb=b[DB_serial]; *bb == '0'; bb++);
        return(strcmp(aa,bb));
        }

static unsigned long index_name_hash(char **a)
        { return(lh_strhash(a[DB_name])); }

static int index_name_qual(char **a)
        { return(a[0][0] == 'V'); }

static int index_name_cmp(char **a, char **b)
        { return(strcmp(a[DB_name],
             b[DB_name])); }


/*
** Functions: main and tool func to operate the OCSP requests/response
** ===================================================================
*/

static int MS_CALLBACK key_callback(char *buf, int len, int verify, void *u)
        {
        int i;

        if (key == NULL) return(0);
        i=strlen(key);
        i=(i > len)?len:i;
        memcpy(buf,key,i);
        return(i);
        }

static int stringToTBL ( char *text, TBL table[], int elements ) {

	int i;
	TBL *elem;

	if( text == NULL ) return(0);

	for( i = 0; i < elements; i++ ) {
		elem = &table[i];
		if( strcmp(elem->name, text) == 0 )
			return elem->type;
	}

	/* if it is not found, we return a zero value */
	return 0;

}

#undef BSIZE
#define BSIZE 256

int main ( int argc, char *argv[] )
	{
	int badops = 0;
	int ret = 1;
	int i;
	BIO *in = NULL;
	BIO *out = NULL;
	BIO *Sin = NULL;
	BIO *Sout = NULL;
	BIO *keyf = NULL;
	BIO *certf = NULL;
	BIO *cacertf = NULL;

	int verify = 0;
	int sign = 0;
	int setmode = 0;
	int new = 0;

	char *ocspTask = NULL;
	char *inform = NULL;
	char *outform = NULL;
	char *infile = NULL;
	char *outfile = NULL;
	char *keyfile = NULL;
	char *configfile = NULL;
	char *dbfile = NULL;
	char *cert = NULL;
	char *cacert = NULL;
	char *certsfile = NULL;
	char *format = NULL;
	char **pp = NULL;
	char *prog = NULL;
	char *tmp = NULL;
	char *ser = NULL;

        MS_STATIC char buf[3][BSIZE];

	char *ocsp_digest_name = NULL;
	EVP_MD *ocsp_digest = NULL;

	TBL **table = NULL;

	LHASH *conf;
	EVP_PKEY *pkey = NULL;
	TXT_DB *db=NULL;
	long errorline= -1;

	X509 *X509_cert;
	X509 *X509_stat;
	X509 *X509_cacert;

	X509V3_add_standard_extensions();
	/* SSLeay_add_all_algorithms(); */
	OpenSSL_add_all_algorithms();

	if (bio_err == NULL)
		if ((bio_err=BIO_new(BIO_s_file())) != NULL)
			BIO_set_fp(bio_err,stderr,BIO_NOCLOSE|BIO_FP_TEXT);

	prog = argv[0];
	if( argc > 1 ) { 
		argv++;
		argc--;
		ocspTask = *argv;
		if( stringToTBL(ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
						 == OCSP_NONE ) { 
			BIO_printf(bio_err,"ERROR: missing <command>\n");
			badops=1;
		}
		if( stringToTBL(ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
				 == OCSP_STATUS_CMD ) {
			if (--argc < 1) goto bad;
			++argv;
			argc--;

			/* We need a even len */
			ser = OPENSSL_malloc( strlen(*argv) + 2);
			if( strlen( *argv ) % 2 ) {
				strcpy( ser, "0" );
				strcat( ser, *argv );
			} else {
				strcpy( ser, *argv );
			};
		}
		/* Let get the Certificate from the -in or STDIN files */
		/*
		if( (stringToTBL(ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
				 == OCSP_SUSPEND_CMD ) or 
		    (stringToTBL(ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
				 == OCSP_UNSUSPEND_CMD )) { 
			if (--argc < 1) goto bad;
			certfile=*(argv++);
			argc--;

		}
		*/
		if( stringToTBL(ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
						 == OCSP_VERSION_CMD ) { 
			BIO_printf(bio_err,"ocsp command version %s\n", BUILD_INFO[0]);
			goto end;
		}
	} else {
		BIO_printf(bio_err,"ERROR: missing <command>\n");
		badops=1;
	}

	argv++;
	argc--;

	while (argc >= 1)
		{
		/* Debuggin options */
#ifdef DEBUG
		BIO_printf(bio_err,"Parsing (%s)\n", *argv);
#endif
		if (strcmp(*argv,"-in") == 0)
			{
			if (--argc < 1) goto bad;
			infile= *(++argv);
			}
		else if (strcmp(*argv,"-out") == 0)
			{
			if (--argc < 1) goto bad;
			outfile= *(++argv);
			}
                else if (strcmp(*argv,"-config") == 0)
                        {
                        if (--argc < 1) goto bad;
                        configfile= *(++argv);
                        }
		else if (strcmp(*argv,"-key") == 0)
			{
			if (--argc < 1) goto bad;
			key= *(++argv);
			}
		else if (strcmp(*argv,"-keyfile") == 0)
			{
			if (--argc < 1) goto bad;
			keyfile= *(++argv);
			}
		else if (strcmp(*argv,"-cert") == 0)
			{
			if (--argc < 1) goto bad;
			cert= *(++argv);
			}
		else if (strcmp(*argv,"-cacert") == 0)
			{
			if (--argc < 1) goto bad;
			cacert= *(++argv);
			}
		else if (strcmp(*argv,"-certs") == 0)
			{
			if (--argc < 1) goto bad;
			certsfile= *(++argv);
			}
		else if (strcmp(*argv,"-inform") == 0)
			{
			if (--argc < 1) goto bad;
			inform= *(++argv);
			}
		else if (strcmp(*argv,"-outform") == 0)
			{
			if (--argc < 1) goto bad;
			outform= *(++argv);
			}
		else if (strcmp(*argv,"-digest") == 0)
			{
			if (--argc < 1) goto bad;
			ocsp_digest_name = *(++argv);
			}
		else if (strcmp(*argv,"-sign") == 0)
			sign=1;
		else if (strcmp(*argv,"-new") == 0)
			new=1;
		else if (strcmp(*argv,"-verify") == 0)
			verify=1;
		else if (strcmp(*argv,"-verbose") == 0)
			verbose=1;
		else badops = 1;
		argc--;
		argv++;
		}

bad:
	if (badops) {
		
		for (pp=ocsp_usage; (*pp != NULL); pp++)
			BIO_printf(bio_err,*pp);
		goto err;
	}

	ERR_load_crypto_strings();

	in=BIO_new(BIO_s_file());
	out=BIO_new(BIO_s_file());
	Sout=BIO_new(BIO_s_file());

	if ((in == NULL) || (out == NULL) || (Sout == NULL)) {
		BIO_printf(bio_err,"Errors creating I/O files!");
		ERR_print_errors(bio_err);
		goto err;
	}
	/*****************************************************************/
        /* we need to load the configuration file */

        if (configfile == NULL) configfile = getenv("OCSP_CONF");
        if (configfile == NULL) goto err;

        BIO_printf(bio_err,"Using configuration from %s\n",configfile);
        if ((conf=CONF_load(NULL,configfile,&errorline)) == NULL)
                {
                if (errorline <= 0)
                        BIO_printf(bio_err,
				"error loading the config file '%s'\n",
                                configfile);
                else
                        BIO_printf(bio_err,
				"error on line %ld of config file '%s'\n"
                                ,errorline, configfile);
                goto err;
                }

        if (section == NULL)
                {
                section=CONF_get_string(conf,BASE_SECTION,ENV_DEFAULT_OCSP);
                if (section == NULL)
                        {
                        lookup_fail(BASE_SECTION,ENV_DEFAULT_OCSP);
                        goto err;
                        }
                }

	/*****************************************************************/
	/* we need to load the db file */
        if ((dbfile=CONF_get_string(conf,section,ENV_DATABASE)) == NULL)
                {
                lookup_fail(section,ENV_DATABASE);
                goto err;
                }

        if (BIO_read_filename(in,dbfile) <= 0)
                {
                perror(dbfile);
                BIO_printf(bio_err,"unable to open '%s'\n",dbfile);
                goto err;
                }
        db=TXT_DB_read(in,DB_NUMBER);
        if (db == NULL) goto err;

        if (!TXT_DB_create_index(db,DB_serial,NULL,index_serial_hash,
                index_serial_cmp))
                {
                BIO_printf(bio_err,
			"error creating serial number index:(%ld,%ld,%ld)\n",
			db->error, db->arg1, db->arg2);
                goto err;
                }

        if (!TXT_DB_create_index(db,DB_name,index_name_qual,index_name_hash,
                index_name_cmp))
                {
                BIO_printf(bio_err,"error creating name index:(%ld,%ld,%ld)\n",
                        db->error,db->arg1,db->arg2);
                goto err;
		}

	/*****************************************************************/
	/* Load private OCSP key */
        if ((keyfile == NULL) && ((keyfile=CONF_get_string(conf,
                section,ENV_PRIVATE_KEY)) == NULL))
                {
		if( verbose ) lookup_fail(section,ENV_PRIVATE_KEY);
                }
	if( keyfile )
		{
		if( verbose )
			BIO_printf( bio_err,
				"Loading key file (%s)\n", keyfile );

		if (BIO_read_filename(in,keyfile) <= 0)
                	{
                	perror(cert);
                	BIO_printf(bio_err,"opening key file\n");
                	goto err;
                	}
		if( key == NULL )
			{
			/* No password required for private key */
			pkey = PEM_read_bio_PrivateKey(in, NULL, NULL, NULL);
			}
		 else
			{
			/* What a mess ... needed to declare key *static* :0 */
			pkey = PEM_read_bio_PrivateKey(in, NULL, key_callback,
				NULL);

			/* Let's delete the private key's passwd from mem */
			memset(key,0,strlen(key));
			}

		if( pkey == NULL )
			{
			BIO_printf(bio_err, "Error reading private key\n");
			goto end;
			}

		}

	/*****************************************************************/
	/* Load the OCSP certificate file */
        if ((cert == NULL) && ((cert=CONF_get_string(conf,
                section,ENV_OCSP_CERTIFICATE)) == NULL))
                {
                if( verbose ) lookup_fail(section,ENV_OCSP_CERTIFICATE);
                }
	if ( cert )
		{
        	if (BIO_read_filename(in,cert) <= 0)
                	{
                	perror(cert);
                	BIO_printf(bio_err,"trying to load OCSP certificate\n");
                	goto err;
                	}
        	X509_cert = PEM_read_bio_X509(in,NULL,NULL,NULL);
        	if (X509_cert == NULL)
                	{
                	BIO_printf(bio_err,"unable to load OCSP certificate\n");
                	goto err;
                	}
        	if (!X509_check_private_key(X509_cert,pkey))
                	{
                	BIO_printf(bio_err,
				"OCSP cert and private key do not match\n");
                	goto err;
                	}
		}

	/*****************************************************************/
	/* Load the CA certificate file */
        if ((cacert == NULL) && ((cacert=CONF_get_string(conf,
                section,ENV_CA_CERTIFICATE)) == NULL))
                {
                lookup_fail(section,ENV_CA_CERTIFICATE);
                goto err;
                }
        if (BIO_read_filename(in,cacert) <= 0)
                {
                perror(cacert);
                BIO_printf(bio_err,"trying to load CA certificate\n");
                goto err;
                }
        X509_cacert = PEM_read_bio_X509(in,NULL,NULL,NULL);
        if (X509_cacert == NULL)
                {
                BIO_printf(bio_err,"unable to load CA certificate\n");
                goto err;
                }


	/*****************************************************************/
	/* Open Input/Output file if given, otherwise use stdin/out */
	if (infile)
		in = BIO_new_file(infile, "r");
	else 
		in = BIO_new_fp(stdin, BIO_NOCLOSE);

	if( in == NULL ) {
		BIO_printf(bio_err, "Error opening input file\n");
		ERR_print_errors(bio_err);
		goto end;
	}

	if (outfile)
		out = BIO_new_file(outfile, "w");
	else
		out = BIO_new_fp(stdout, BIO_NOCLOSE);

	if(!out) {
		BIO_printf(bio_err, "Error opening output file\n");
		ERR_print_errors(bio_err);
		goto end;
	}

	/*****************************************************************/
	/* Let's get the digest */
        if ((ocsp_digest_name == NULL) &&
		((ocsp_digest_name=CONF_get_string(conf,section,
						   ENV_DEFAULT_MD)) == NULL))
                {
                lookup_fail(section,ENV_DEFAULT_MD);
                goto err;
                }

	ocsp_digest = (EVP_MD *) EVP_get_digestbyname(ocsp_digest_name);
	if( ocsp_digest == NULL )
		{
		BIO_printf(bio_err, "ERROR: unsupported digest type %s\n",
							ocsp_digest_name);
                goto err;
                }
	
	/*****************************************************************/
	/* Let's get the Certificate 'ser' status */
	if( stringToTBL( ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
							 == OCSP_STATUS_CMD ) {
		char 		status;
		ASN1_INTEGER	*a = NULL;
		BIO		*tmpSer;

		if( ! ser ) {
			BIO_printf( bio_err,
				"ERROR: missing serial number!" );
			goto err;
		};

		a = ASN1_INTEGER_new();
		if( a == NULL ) goto err;
		if ((tmpSer=BIO_new(BIO_s_mem())) == NULL)
			goto err;

		(void) BIO_reset(tmpSer);
		if (!BIO_puts(tmpSer, ser)) {
			BIO_free(tmpSer);
			goto err;
		}

		if (!a2i_ASN1_INTEGER(tmpSer,a,buf[0],BSIZE))
			{
			BIO_free(tmpSer);
			BIO_printf(bio_err,"unable to get serial %s\n", ser);
			goto err;
			}

		status = get_status( a, db );
		if( status == DB_TYPE_VAL )
			{
			strncpy( buf[0], "Valid", BSIZE );
			buf[0][5] = '\x0';
			}
	       	else if (status == DB_TYPE_SUS )
			{
			strncpy( buf[0], "Suspended", BSIZE );
			buf[0][9] = '\x0';
			}
	       	else if (status == DB_TYPE_REV )
			{
			strncpy( buf[0], "Revoked", BSIZE );
			buf[0][7] = '\x0';
			}
	       	else
			{
			strncpy( buf[0], "Unknown", BSIZE );
			buf[0][7] = '\x0';
			}
		BIO_printf( out, "Certificate %s : %s (%c)\n",
				ser, buf[0], status );
		BIO_free(tmpSer);
	}

	/*****************************************************************/
	/* If we want to suspend unsuspend a certificate */
	if( ( stringToTBL( ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
			       	== OCSP_SUSPEND_CMD ) ||
	  ( stringToTBL( ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
	   			== OCSP_UNSUSPEND_CMD )) {
		int cmd;
		char mode;

		BIO *bio_certs;

		if(stringToTBL( ocspTask, ocspCommandsTbl,OCSP_COMMANDS_SIZE )
				 == OCSP_SUSPEND_CMD ) {
			cmd = 0;
			mode = DB_TYPE_SUS;
			}
		else
			{
			cmd =1;
			mode = DB_TYPE_VAL;
			};

		if( verbose ) {
			if( cmd == 0 )
				BIO_printf( bio_err, "Suspending Certificate" );
			else
				BIO_printf( bio_err, "Re-Enabling Certificate" );
		}

		if (certsfile) {
			/* Read certificates from certsfile */
			bio_certs = BIO_new(BIO_s_file());
			bio_certs = BIO_new_file( certsfile, "r" );
		} else {
			/* Else we can send error or read from in file */
			bio_certs = in;
		}

		X509_stat = PEM_read_bio_X509(bio_certs,NULL,NULL,NULL);
        	if (X509_stat == NULL)
                	{
                	BIO_printf(bio_err,
				"ERROR: unable to load certificate\n");
			if( certsfile ) BIO_free(bio_certs);
                	goto err;
                	}
		while ( X509_stat ) {
        		if (X509_stat == NULL) break;
			if( set_status( X509_stat, db, mode) != 1 ) {
				BIO_printf( bio_err,
					"ERROR: setting mode %c\n",
			       		mode);
				if( certsfile ) BIO_free(bio_certs);
				goto err;
			}

			X509_stat = PEM_read_bio_X509( bio_certs, 
					NULL,NULL, NULL );
		}
		
		X509_free(X509_stat);
		if( certsfile ) BIO_free(bio_certs);

		strncpy(buf[0],dbfile,BSIZE-4);
		strcat(buf[0],".new");
		if (BIO_write_filename(out,buf[0]) <= 0)
			{
			perror(dbfile);
			BIO_printf(bio_err,"unable to open '%s'\n",dbfile);
			goto err;
			}
		if( TXT_DB_write(out,db) <= 0 ) goto err;

		strncpy(buf[1],dbfile,BSIZE-4);
		strcat(buf[1],".old");
		if (rename(dbfile,buf[1]) < 0)
			{
			BIO_printf(bio_err,"unable to rename %s to %s\n",
					dbfile, buf[1]);
			perror("reason");
			goto err;
			}
		if (rename(buf[0],dbfile) < 0)
			{
			BIO_printf(bio_err,"unable to rename %s to %s\n",
					buf[0],dbfile);
			perror("reason");
			rename(buf[1],dbfile);
			goto err;
			}
		BIO_printf(bio_err,"Data Base Updated\n"); 
	}
	/*****************************************************************/
	/* if we need a new request|response */
	if( new ) {
		if(verbose) {
			BIO_printf(bio_err,"Generating new %s\n", ocspTask );
			BIO_printf(bio_err,"Using digest type: %s\n",
					   ocsp_digest_name );
		}

		if( stringToTBL( ocspTask, ocspCommandsTbl, OCSP_COMMANDS_SIZE )
							 == OCSP_REQUEST_CMD ) {
			OCSP_REQUEST *request = NULL;

			request = generateNewRequest( X509_cert,
				 X509_cacert, pkey, ocsp_digest,
				 certsfile, sign );

			if( request == NULL )
				{
				BIO_printf( bio_err,
					"ERROR: can not create new request.\n");
				goto err;
				}

			if( stringToTBL( outform, ocspFormatsTbl,
				 	OCSP_FORMAT_SIZE) == OCSP_ASN1_FORMAT )
				{
				i2d_OCSP_REQ_bio( out, request);
				}

			if( stringToTBL( outform, ocspFormatsTbl,
				 	OCSP_FORMAT_SIZE) == OCSP_TXT_FORMAT ){
				BIO_printf( out, "Request:\n");
				printRequest( out, request );
			}
		}
	}

err:

end:
	if(in) BIO_free(in);
	if(out) BIO_free(out);
	if(keyf) BIO_free(keyf);
	if(pkey) EVP_PKEY_free(pkey);
	EXIT(ret);
	}


/* Allocation Functions */
OCSP_REQINFO *OCSP_REQINFO_new() {
	OCSP_REQINFO *ret = NULL;

	ret = (OCSP_REQINFO *) OPENSSL_malloc (sizeof(OCSP_REQINFO));
	if( ret ) {
		/* If allocation went well, we want to erase
		   allocked memory */
		ret->version = NULL;
		ret->requestorName = NULL;
		ret->requestList = NULL;
		ret->requestExtensions = NULL;
	}

	return ret;
}


/* Generate a new OCSP Request */
OCSP_REQUEST *generateNewRequest( X509 *cert, X509 *cacert, EVP_PKEY *pkey,
	       			EVP_MD *ocsp_digest, char *certsfile,
			       	int sign ) {
	OCSP_REQUEST 	*ret = NULL;
	X509_NAME 	*name = NULL;
	BIO 		*bio_certs = NULL;
	OCSP_CERTID 	*cid;
        X509 		*c1 = NULL;
        X509 		*c2 = NULL;
	char 		buf[256];
	STACK_OF(X509)	*sk = NULL;

	if ((ret=(OCSP_REQUEST *)OPENSSL_malloc(sizeof(OCSP_REQUEST)))==NULL )
		goto err;

	/* Let's allocate Memory for the REQINFO structure */
	ret->tbsRequest = OCSP_REQINFO_new();
	ret->optionalSignature = NULL;

	/* Get the Requestor Subject Name from passed certificate */
	if( cert )
		name = X509_get_subject_name(cert);

	if (name)
        {
		if(verbose) {
			X509_NAME_oneline(name,buf,256);
			BIO_printf(bio_err,"Certificate Subject: %s\n", buf );
		}

                if (!(ret->tbsRequest->requestorName=GENERAL_NAME_new()))
                        goto err;
                ret->tbsRequest->requestorName->type = GEN_DIRNAME;
                ret->tbsRequest->requestorName->d.dirn = X509_NAME_dup(name);
        }

        if (!(ret->tbsRequest->requestList = sk_new(NULL)))
		goto err;

	if (certsfile)
                {

		/* Read certificates from certsfile */
		bio_certs = BIO_new(BIO_s_file());
		bio_certs = BIO_new_file( certsfile, "r" );

		c1 = PEM_read_bio_X509( bio_certs, NULL,NULL, NULL );
		while ( c1 ) {
			c2 = X509_dup( c1 );

			if (!(cid = OCSP_cert_id_new(ocsp_digest,
                                             X509_get_issuer_name(c2),
                                             c2->cert_info->key->public_key,
                                             X509_get_serialNumber(c2))))
                        	goto err;

			if (!(OCSP_request_add(ret, cid, NULL))) {
				BIO_printf(bio_err,"ERROR: error adding request\n");
				goto err;
			}

			c1 = PEM_read_bio_X509( bio_certs, NULL,NULL, NULL );
		}
		
	}

	if( sign ) {
        	if (!(sk = sk_new(NULL))) goto err;
		if (cert && !sk_push(sk, (char*) cert)) goto err;
		if (cacert && !sk_push(sk, (char*) cacert)) goto err;
		OCSP_request_sign( ret, pkey, ocsp_digest, sk );
	}

	return(ret);

err:
	if(!ret) {
		BIO_printf(bio_err,"ERROR: memory allocation error(RET)\n");
	}
	if(!cid) {
		BIO_printf(bio_err,"ERROR: memory allocation error(CID)\n");
	}

	return;
}

int OCSP_request_sign(OCSP_REQUEST *req, EVP_PKEY *key, EVP_MD *dgst,
		      STACK_OF(X509) *certs) {
	int i;
	OCSP_SIGNATURE *sig;

	/* Get an allocked signature structure */
	if (!(req->optionalSignature = sig = OCSP_SIGNATURE_new()))
		goto err;

	/* Let's sign the req->tbsRequest bytes */
	if( ! ASN1_sign((int(*)())i2d_OCSP_REQINFO,
			req->optionalSignature->signatureAlgorithm,
			NULL, sig->signature,
			(char *) req->tbsRequest,key,dgst))
		goto err;

	/* If there are certs, let's add them to the request,
	 * usually them will be requestor's and ca's certs */
	if (certs)
	        {
	        if (!(sig->certs = sk_dup(certs)))
			goto err;
	        for (i = 0; i < sk_num(sig->certs); i++)
	                {
		        if (!(sig->certs->data[i] = (char*)
		               X509_dup((X509*)certs->data[i]))) goto err;
		        }
		}
	return 1;
err:
	if (req->optionalSignature)
	        {
		OCSP_SIGNATURE_free(req->optionalSignature);
		req->optionalSignature = NULL;
		}
	return 0;
	}

OCSP_SIGNATURE *OCSP_SIGNATURE_new(void) {
	OCSP_SIGNATURE *ret;

	if((ret = OPENSSL_malloc( sizeof( OCSP_SIGNATURE ))) == NULL )
		goto err;

	if((ret->signatureAlgorithm = X509_ALGOR_new()) == NULL ) {
		goto err;
	}

	if ((ret->signature = (ASN1_BIT_STRING*) ASN1_OBJECT_new()) == NULL) {
		goto err;
	}

	ret->certs 	= NULL;
	
	return (ret);

err:
	return(NULL);
}

int OCSP_SIGNATURE_free( OCSP_SIGNATURE *sig ) {

	/* To be ended, it lacks signature_free and sig_alg_free */
	if( sig->certs )
		sk_pop_free( sig->certs, (void(*)()) X509_free);
	sig->certs = NULL;

	return(1);
}


int printRequest(BIO *bp, OCSP_REQUEST *o) {
        int i,j,n;
        long l;
        char *s;
        OCSP_CERTID* cid = NULL;
        ASN1_INTEGER *ai = NULL;
        OCSP_ONEREQ *one = NULL;
        OCSP_REQINFO *inf = o->tbsRequest;
        OCSP_SIGNATURE *sig = o->optionalSignature;

        if (BIO_write(bp,"OCSP Request Data:\n",19) <= 0)
		goto err;

        l=ASN1_INTEGER_get(inf->version);
        if (BIO_printf(bp,"%4sVersion: %lu (0x%lx)","",l+1,l) <= 0)
		goto err;

        if (inf->requestorName != NULL)
        {
                if (BIO_write(bp,"\n    Requestor Name: ",21) <= 0) 
                        goto err;
                genName (bp, inf->requestorName);
        }

        if (BIO_write(bp,"\n    Requestor List:\n",21) <= 0)
		goto err;

        for (i = 0; i < sk_num(inf->requestList); i++)
        {
                if (!inf->requestList->data[i])
			continue;
                one = (OCSP_ONEREQ*) inf->requestList->data[i];
                cid = one->reqCert;
                j=OBJ_obj2nid(cid->hashAlgorithm->algorithm);
                if (BIO_write(bp,"        Serial Number: ",23) <= 0) 
                        goto err;
                if (!i2a_ASN1_INTEGER(bp, cid->serialNumber)) 
                        goto err;
                if (BIO_printf(bp,"\n%12sHash Algorithm: %s","",
                               (j == NID_undef)?"UNKNOWN":OBJ_nid2ln(j)) <= 0)
                        goto err;
                if (BIO_write(bp,"\n            Issuer Name Hash: ",27) <= 0)
                        goto err;
                i2a_ASN1_STRING(bp, cid->issuerNameHash, V_ASN1_OCTET_STRING);
                if (BIO_write(bp,"\n            Issuer Key Hash: ",26) <= 0) 
                        goto err;
                i2a_ASN1_STRING(bp, cid->issuerKeyHash, V_ASN1_OCTET_STRING);
                if (!BIO_write(bp,"\n",1)) goto err;
		/*
                if (!OCSP_extensions_print(bp, one->singleRequestExtensions,
                                           "Request Single Extensions"))
                        goto err;
		*/
        }
/*
        if (!OCSP_extensions_print(bp, inf->requestExtensions, 
                                   "Request Extensions"))
                goto err;
*/
        if (sig)
                {

                i=OBJ_obj2nid(sig->signatureAlgorithm->algorithm);
                if (BIO_printf(bp,"    Signature Algorithm: %s",
                               (i == NID_undef)?"UNKNOWN":OBJ_nid2ln(i)) <= 0)
                        goto err;
		
                n=sig->signature->length;
                s=(char *)sig->signature->data;
                for (i=0; i<n; i++)
                        {
                        if ((i%18) == 0)
                        if (BIO_write(bp,"\n        ",9) <= 0) goto err;
                        if (BIO_printf(bp,"%02x%s",(unsigned char)s[i],
                                ((i+1) == n)?"":":") <= 0) goto err;
                        }
                if (BIO_write(bp,"\n",1) != 1) goto err;
                if (sig->certs)
                        {
                        for (i=0; i<sig->certs->num; i++)
                                if (sig->certs->data[i] != NULL)
                                        X509_print(bp,
                                                   (X509*)sig->certs->data[i]);
                        }
                }
        return 1;
err:
        return 0;
        }


int genName(BIO *bp, GENERAL_NAME *n)
        {
        int j;
        char *p;

        if (n == NULL) return(0);

        switch (n->type)
                {

        case GEN_DIRNAME:
                X509_NAME_print(bp,n->d.dirn,16);
                break;

        case GEN_EMAIL:
        case GEN_DNS:
        case GEN_URI:
        case GEN_IPADD:
                p=(char *)n->d.ip->data;
                for (j=n->d.ip->length;j>0;j--)
                        {
                        if ((*p >= ' ') && (*p <= '~'))
                                BIO_printf(bp,"%c",*p);
                        else if (*p & 0x80)
                                BIO_printf(bp,"\\0x%02X",*p);
                        else if ((unsigned char)*p == 0xf7)
                                BIO_printf(bp,"^?");
                        else    BIO_printf(bp,"^%c",*p+'@');
                        p++;
                        }
                break;

        case GEN_RID:
                i2a_ASN1_OBJECT(bp, n->d.rid);
                break;

        /* XXX these are legit, need to support at some time... */
        case GEN_OTHERNAME:
        case GEN_X400:
        case GEN_EDIPARTY:
        default:
                return 0;
                }

        return 1;
        }

OCSP_CERTID *OCSP_cert_id_new(EVP_MD *dgst, X509_NAME *issuerName, 
		ASN1_BIT_STRING* issuerKey, ASN1_INTEGER *serialNumber) {
        int 		nid;
        unsigned int 	i;
        X509_ALGOR 	*alg;
        OCSP_CERTID 	*cid = NULL;
        unsigned char 	md[EVP_MAX_MD_SIZE];

	/* Allocate the memory for the CID structure */
        if (!(cid = (OCSP_CERTID *) OPENSSL_malloc (sizeof(OCSP_CERTID)) ))
		goto err;

	/* Get the X509 Algorithm used in the certificate */
	if((alg = X509_ALGOR_new()) == NULL ) {
		goto err;
	}

	/* Get the digest algorithm used */
        if((nid = EVP_MD_type(dgst)) == NID_undef ) {
                goto err;
	}

	/* Set the CID hash algorithm */
	cid->hashAlgorithm = alg;

	/* If cannot set the algorithm, then error */
        if (!(alg->algorithm=OBJ_nid2obj(nid)))
		goto err;
        if ((alg->parameter=ASN1_TYPE_new()) == NULL)
		goto err;
        alg->parameter->type=V_ASN1_NULL;

        if (!X509_NAME_digest(issuerName, dgst, md, &i)) goto digerr;
        if (!(cid->issuerNameHash = ASN1_OCTET_STRING_new())) goto err;
        if (!(ASN1_OCTET_STRING_set(cid->issuerNameHash, md, i))) goto err;


	if( ! ASN1_digest((int (*)())i2d_ASN1_BIT_STRING,
				dgst,(char *)issuerKey,md,&i))
			goto digerr;

        if (!(cid->issuerKeyHash = ASN1_OCTET_STRING_new())) goto err;
        if (!(ASN1_OCTET_STRING_set(cid->issuerKeyHash, md, i))) goto err;
        
	/*
        if (cid->serialNumber != NULL) ASN1_INTEGER_free(cid->serialNumber);
	*/

	if (!(cid->serialNumber = ASN1_INTEGER_dup(serialNumber))) goto err;

        return cid;
digerr:
        /* OCSPerr(OCSP_F_CERT_ID_NEW,OCSP_R_DIGEST_ERR); */
	BIO_printf(bio_err,"ERROR: hash algorithm not supported\n");
err:
        /* if (cid) OCSP_CERTID_free(cid); */
	BIO_printf(bio_err,"ERROR: creating cert id structure\n");
        return NULL;
        }


int OCSP_request_add(OCSP_REQUEST *req, OCSP_CERTID *cid, STACK *extensions)
        {
        OCSP_ONEREQ *one = NULL;

        if (!(one = (OCSP_ONEREQ *) OPENSSL_malloc(sizeof(OCSP_ONEREQ))))
		goto err;

/*
	one->reqCert = (OCSP_CERTID *)ASN1_dup((int(*)())i2d_OCSP_CERTID,
			 (char *(*)())d2i_OCSP_CERTID,(char *)(cid));
*/
	
	one->reqCert = cid;

        if (!(one->reqCert))
		goto err;
/*
        if (extensions&&(!(one->singleRequestExtensions=ext_dup(extensions))))
                goto err;
*/
        if (!sk_push(req->tbsRequest->requestList,(char*)one)) {
		BIO_printf( bio_err, "ERROR: pushing request to requestList\n");
		goto err;
	}

        return 1;
err:
        return 0;
}

STACK_OF(X509) *read_certs(char *file, int min)
        {
        int i;
        FILE *fp;
        X509* c1, *c2;
        STACK_OF(X509) *sk1, *sk2 = NULL;
        PKCS7 *pk7 = NULL;

        if ((fp = fopen(file, "rb")) == NULL) goto err;
        pk7 = PEM_read_PKCS7(fp, NULL, NULL, NULL);
        fclose(fp);
        fp = NULL;
        if (!pk7)
                {
                if ((fp = fopen(file, "rb")) == NULL) goto err;
                c1 = c2 = PEM_read_X509(fp, NULL, NULL, NULL );
                fclose(fp);
                fp = NULL;
                if (!c1)
                        {
                        BIO_printf(bio_err,"failed to read cert in %s\n",file);
                        goto err;
                        }
                if (!(sk1 = sk_new(NULL))) goto err;
                if (!sk_push(sk1, (char*) c1)) goto err;
                }
        else
                {
                if (!PKCS7_type_is_signed(pk7)) goto err;
                sk1 = (STACK_OF(X509) *) pk7->d.sign->cert;
                if (sk_num(sk1) < min) goto err;
                }
        c1 = c2 = (X509*) sk1->data[sk_num(sk1)-1];
        if (min > 1)
                {
                if (sk_num(sk1) > 1) c2 = (X509*) sk1->data[sk_num(sk1)-2];
                if (X509_name_cmp(X509_get_issuer_name(c1),
                            X509_get_subject_name(c2)))
                        {
                        BIO_printf(bio_err, "issuer name of user cert ");
                        X509_NAME_print(bio_err, X509_get_issuer_name(c1),16);
                        BIO_printf(bio_err,
                              " is not same as subject name of issuer cert  ");
                        X509_NAME_print(bio_err, X509_get_subject_name(c2),16);
                        BIO_write(bio_err, "\n", 1);
                        goto err;
                        }
                }
        if (!(sk2 = sk_dup(sk1))) goto err;
        for (i = 0; i < sk_num(sk1); i++)
                {
                if (!(sk2->data[i] = (char*)
                               X509_dup((X509*)sk1->data[i]))) goto err;
                }
        if (pk7) PKCS7_free(pk7);
        else sk_pop_free(sk1, (void(*)()) X509_free);
        return sk2;
err:
        if (pk7) PKCS7_free(pk7);
        if (sk2) sk_pop_free(sk2, (void(*)()) X509_free);
        return 0;
        }



int i2d_OCSP_REQ ( OCSP_REQUEST *a, unsigned char **pp) {
	/*
	** ASN1 functions (hope this works fine ... please give feedback,
	** I am not at ease with this OpenSSL functions ... :-D
	*/

	int num=0;

	M_ASN1_I2D_vars(a);

        M_ASN1_I2D_len(a->tbsRequest, i2d_OCSP_REQINFO);
        M_ASN1_I2D_len_EXP_opt(a->optionalSignature, i2d_OCSP_SIGNATURE,0,num);
        M_ASN1_I2D_seq_total();

#ifdef DEBUG
		BIO_printf( bio_err,"::: seq_total()\n");
#endif

        M_ASN1_I2D_put(a->tbsRequest, i2d_OCSP_REQINFO);
        M_ASN1_I2D_put_EXP_opt(a->optionalSignature, i2d_OCSP_SIGNATURE,0,num);
        M_ASN1_I2D_finish();

}

int i2d_OCSP_REQINFO(OCSP_REQINFO *a, unsigned char **pp)
        {
        int v1=0,v2=0,v3=0;
        M_ASN1_I2D_vars(a);

        M_ASN1_I2D_len_EXP_opt(a->version,i2d_ASN1_INTEGER,0,v1);
        M_ASN1_I2D_len_EXP_opt(a->requestorName,i2d_GENERAL_NAME,1,v2);
        M_ASN1_I2D_len_SEQUENCE(a->requestList,i2d_OCSP_ONEREQ);

/*
        M_ASN1_I2D_len_EXP_SET_opt(a->requestExtensions,i2d_X509_EXTENSION,2,V_ASN1_SEQUENCE,v3);
*/
        M_ASN1_I2D_seq_total();

        M_ASN1_I2D_put_EXP_opt(a->version,i2d_ASN1_INTEGER,0,v1);
        M_ASN1_I2D_put_EXP_opt(a->requestorName,i2d_GENERAL_NAME,1,v2);
        M_ASN1_I2D_put_SEQUENCE(a->requestList,i2d_OCSP_ONEREQ);

/*
        M_ASN1_I2D_put_EXP_SET_opt(a->requestExtensions,i2d_X509_EXTENSION,2,V_ASN1_SEQUENCE,v3);
*/
        M_ASN1_I2D_finish();
        }


int i2d_OCSP_CERTID( OCSP_CERTID *a, unsigned char **pp)
{
        M_ASN1_I2D_vars(a);

        M_ASN1_I2D_len(a->hashAlgorithm, i2d_X509_ALGOR);
        M_ASN1_I2D_len(a->issuerNameHash, i2d_ASN1_OCTET_STRING);
        M_ASN1_I2D_len(a->issuerKeyHash, i2d_ASN1_OCTET_STRING);
        M_ASN1_I2D_len(a->serialNumber, i2d_ASN1_INTEGER);
        M_ASN1_I2D_seq_total();
        M_ASN1_I2D_put(a->hashAlgorithm, i2d_X509_ALGOR);
        M_ASN1_I2D_put(a->issuerNameHash, i2d_ASN1_OCTET_STRING);
        M_ASN1_I2D_put(a->issuerKeyHash, i2d_ASN1_OCTET_STRING);
        M_ASN1_I2D_put(a->serialNumber, i2d_ASN1_INTEGER);
        M_ASN1_I2D_finish();
}

int i2d_OCSP_SIGNATURE(OCSP_SIGNATURE *a,unsigned char **pp) {
        int v=0;

        M_ASN1_I2D_vars(a);

        M_ASN1_I2D_len(a->signatureAlgorithm, i2d_X509_ALGOR);
        M_ASN1_I2D_len(a->signature, i2d_ASN1_BIT_STRING);
        M_ASN1_I2D_len_EXP_SET_opt(a->certs,i2d_X509,0,V_ASN1_SEQUENCE,v);
        M_ASN1_I2D_seq_total();
        M_ASN1_I2D_put(a->signatureAlgorithm, i2d_X509_ALGOR);
        M_ASN1_I2D_put(a->signature, i2d_ASN1_BIT_STRING);
        M_ASN1_I2D_put_EXP_SET_opt(a->certs,i2d_X509,0,V_ASN1_SEQUENCE,v);
        M_ASN1_I2D_finish();
        }

int i2d_OCSP_ONEREQ(OCSP_ONEREQ *a, unsigned char **pp)
        {
        int v=0;
        M_ASN1_I2D_vars(a);

        M_ASN1_I2D_len(a->reqCert, i2d_OCSP_CERTID);
/*
        M_ASN1_I2D_len_EXP_SET_opt(a->singleRequestExtensions,
                                   i2d_X509_EXTENSION,0,V_ASN1_SEQUENCE,v);
*/
        M_ASN1_I2D_seq_total();
        M_ASN1_I2D_put(a->reqCert, i2d_OCSP_CERTID);
/*
        M_ASN1_I2D_put_EXP_SET_opt(a->singleRequestExtensions,
                                   i2d_X509_EXTENSION,0,V_ASN1_SEQUENCE,v);
*/
        M_ASN1_I2D_finish();
        }

/* Reverse ASN1 funcions */

OCSP_CERTID *d2i_OCSP_CERTID( OCSP_CERTID **a, unsigned char **pp, long length)
        {
        M_ASN1_D2I_vars(a,OCSP_CERTID *,OCSP_CERTID_new);

        M_ASN1_D2I_Init();
        M_ASN1_D2I_start_sequence();
        M_ASN1_D2I_get(ret->hashAlgorithm, d2i_X509_ALGOR);
        M_ASN1_D2I_get(ret->issuerNameHash, d2i_ASN1_OCTET_STRING);
        M_ASN1_D2I_get(ret->issuerKeyHash, d2i_ASN1_OCTET_STRING);
        M_ASN1_D2I_get(ret->serialNumber, d2i_ASN1_INTEGER);
	/*
        M_ASN1_D2I_Finish(a,OCSP_CERTID_free, ASN1_F_D2I_OCSP_CERTID);
	*/
err:
        }

OCSP_CERTID *OCSP_CERTID_new()
        {
        ASN1_CTX c;
        OCSP_CERTID *ret=NULL;

	if(!(ret = (OCSP_CERTID *) OPENSSL_malloc (sizeof(OCSP_CERTID))))
	{
		BIO_printf( bio_err, "ERROR: malloc CERTID_new\n");
		return(NULL);
	}

	/*
	ret->hashAlgorithm = X509_ALGOR_new();
	ret->issuerNameHash = ASN1_OBJECT_new();
	ret->issuerKeyHash = ASN1_OBJECT_new();
	ret->serialNumber = ASN1_OBJECT_new();
	*/

        M_ASN1_New(ret->hashAlgorithm, X509_ALGOR_new);
        M_ASN1_New(ret->issuerNameHash, ASN1_OCTET_STRING_new);
        M_ASN1_New(ret->issuerKeyHash, ASN1_OCTET_STRING_new);
        M_ASN1_New(ret->serialNumber, ASN1_INTEGER_new);

        return(ret);
        }

static char get_status ( ASN1_INTEGER *serial, TXT_DB *db ) {

	char 	*row[DB_NUMBER],**rrow,**irow;
	int 	ok=-1,i;
	char 	ret = DB_TYPE_UNK;

	BIGNUM 	*bn = NULL;

	bn = ASN1_INTEGER_to_BN(serial,NULL);
	row[DB_serial]=BN_bn2hex(bn);
	BN_free(bn);

	if ((row[DB_name] == NULL) || (row[DB_serial] == NULL))
		{
		BIO_printf(bio_err,"Memory allocation failure\n");
		goto err;
		}

	rrow=TXT_DB_get_by_index(db,DB_serial,row);
	if (rrow == NULL)
		{
		return DB_TYPE_UNK;
		}
	else
		{
		ret = rrow[DB_type][0];
		return (ret);
		}

err:
	BIO_printf( bio_err, "ERROR: certificate %ld not present in db.\n",
				ASN1_INTEGER_get(serial));
	return '\x0';

}

static int set_status(X509 *x509, TXT_DB *db, char MODE)
{
	ASN1_UTCTIME *tm=NULL, *revtm=NULL;
	char *row[DB_NUMBER],**rrow,**irow;
	BIGNUM *bn = NULL;
	int ok=-1,i;

	for (i=0; i<DB_NUMBER; i++)
		row[i]=NULL;
	row[DB_name]=X509_NAME_oneline(X509_get_subject_name(x509),NULL,0);
	bn = ASN1_INTEGER_to_BN(X509_get_serialNumber(x509),NULL);
	row[DB_serial]=BN_bn2hex(bn);
	BN_free(bn);

	if ((row[DB_name] == NULL) || (row[DB_serial] == NULL))
		{
		BIO_printf(bio_err,"Memory allocation failure\n");
		goto err;
		}

	/* We have to lookup by serial number because name lookup
	 * skips revoked certs
 	 */

	rrow=TXT_DB_get_by_index(db,DB_serial,row);
	if (rrow == NULL)
		{
		if( verbose )
			BIO_printf(bio_err,"Adding Entry to DB for %s\n",
				       row[DB_name]);

		/* We now just add it to the database */
		row[DB_type]=(char *)OPENSSL_malloc(2);

		tm=X509_get_notAfter(x509);
		row[DB_exp_date]=(char *)OPENSSL_malloc(tm->length+1);
		memcpy(row[DB_exp_date],tm->data,tm->length);
		row[DB_exp_date][tm->length]='\0';

		row[DB_rev_date]=NULL;

		/* row[DB_serial] done already */
		row[DB_file]=(char *)OPENSSL_malloc(8);

		/* row[DB_name] done already */

		if ((row[DB_type] == NULL) || (row[DB_exp_date] == NULL) ||
			(row[DB_file] == NULL))
			{
			BIO_printf(bio_err,"Memory allocation failure\n");
			goto err;
			}
		strcpy(row[DB_file],"unknown");
		row[DB_type][0]='V';
		row[DB_type][1]='\0';

		if ((irow=(char **)OPENSSL_malloc(sizeof(char *)*(DB_NUMBER+1))) == NULL)
			{
			BIO_printf(bio_err,"Memory allocation failure\n");
			goto err;
			}

		for (i=0; i<DB_NUMBER; i++)
			{
			irow[i]=row[i];
			row[i]=NULL;
			}
		irow[DB_NUMBER]=NULL;

		if (!TXT_DB_insert(db,irow))
			{
			BIO_printf(bio_err,"failed to update database\n");
			BIO_printf(bio_err,"TXT_DB error number %ld\n",
					db->error);
			goto err;
			}

		if( verbose )
			BIO_printf(bio_err,"Added Entry to DB %s\n",
				       irow[DB_serial]);

		ok = set_status(x509,db,MODE);

		goto err;

		}
	else if (index_name_cmp(row,rrow))
		{
		BIO_printf(bio_err,"ERROR:name does not match %s\n",
			   row[DB_name]);
		goto err;
		}
	else if (rrow[DB_type][0]== DB_TYPE_REV && MODE == DB_TYPE_SUS)
		{
		BIO_printf(bio_err,"ERROR:Already revoked, serial number %s\n",
			   row[DB_serial]);
		goto err;
		}
	else if (rrow[DB_type][0]== DB_TYPE_EXP && MODE == DB_TYPE_SUS)
		{
		BIO_printf(bio_err,"ERROR:Already expired, serial number %s\n",
			   row[DB_serial]);
		goto err;
		}
	else
		{
		BIO_printf(bio_err,"Updating Certificate %s status.\n",
							rrow[DB_serial]);
		revtm = ASN1_UTCTIME_new();
		revtm=X509_gmtime_adj(revtm,0);

		rrow[DB_type][0]= MODE;
		rrow[DB_type][1]='\0';

		if( MODE == DB_TYPE_SUS ) {
			rrow[DB_rev_date]=(char *)
					OPENSSL_malloc(revtm->length+1);
			memcpy(rrow[DB_rev_date],revtm->data,revtm->length);
			rrow[DB_rev_date][revtm->length]='\0';
		} else {
			rrow[DB_rev_date] = '\0';
		}

		ASN1_UTCTIME_free(revtm);
		}
	ok=1;
err:
	for (i=0; i<DB_NUMBER; i++)
		{
		if (row[i] != NULL) 
			OPENSSL_free(row[i]);
		}
	return(ok);
}

