#!/usr/bin/perl

$|=1;

use OpenCA::OpenSSL;
use OpenCA::X509;
use OpenCA::DB;

my $VER = '0.3.01';

print "\n\nCertificates Import (valid_certs - Version $VER\n";
print "(c) 1999/2000 by Massimiliano Pala and OpenCA Team\n\n";

if( not $ARGV[0] ) {
	print "Usage: importCerts.pl certs-dir [openca-dir] [openssl-path]\n\n";
	print "certs-dir    - Directory where PEM certificates are stored\n";
	print "openca-dir   - OpenCA directory (def. /usr/local/OpenCA)\n";
	print "openssl-path - Path to the openssl command " . 
			      "(def. /usr/local/ssl/bin/openssl\n";
	print "\n";
	exit 0;
}

my $srcDir    = $ARGV[0];

my $opencaDir = ( $ARGV[1] or "/usr/local/OpenCA" );
my $sslPath   = ( $ARGV[2] or "/usr/local/ssl/bin/openssl" );

print " * Initializing crypto shell ... ";
my $openssl = new OpenCA::OpenSSL;

$openssl->setParams(SHELL=>$sslPath,
		    CONFIG=>"$opencaDir/stuff/openssl.cnf",
		    STDERR=>"/dev/null" );

if( not $openssl ) {
	print "Error!\n\n";
	exit 1;
} else {
	print "Ok.\n";
}

print " * Initializing dB object ... ";
my $db = new OpenCA::DB( SHELL=>$openssl, DB_DIR=>"$opencaDir/dB" );

if( not $db ) {
        print "Error!\n\n";
        exit 1;
} else {
	print "Ok.\n\n";
}

opendir( DIR, "$srcDir" ) or die "Error, cannot open $srcDir!\n\n";
	my @list = grep( /[a-f0-9]+\.pem/i , readdir( DIR ) );
closedir( DIR );

foreach $tmpCert ( @list ) {
	my $X509 = new OpenCA::X509( INFILE=>"$srcDir/$tmpCert",
			             FORMAT=>"PEM",
				     SHELL=>$openssl);

	next if ( not $X509 );

	print " * New Certificate Data:\n";
	print "   Serial: " . $X509->getParsed()->{SERIAL} . "\n";
	print "   Version: " . $X509->getParsed()->{VERSION} . "\n";
	print "   C-Name:  " . $X509->getParsed()->{CN} . "\n";
	print "   E-Mail: " . $X509->getParsed()->{EMAIL} . "\n\n";

	print "Storing Certificate to dB ... ";
	if( not $db->storeItem( DATATYPE=>VALID_CERTIFICATE,
				DATA=>$X509->getPEM() ) ) {
  		print "Error!\n\n";
   		next;
	} else {
		print "Ok.\n\n";
		unlink( "$srcDir/$tmpCert" );
	}

}

print "\n * New dB Status:\n\n";
print "   STATUS   => " . $db->getItem( DATATYPE =>VALID_CERTIFICATE,
				     KEY=>STATUS ) . "\n";
print "   INIT     => " . $db->getItem( DATATYPE =>VALID_CERTIFICATE,
				     KEY=>INIT ) . "\n";
print "   MODIFIED => " . $db->getItem( DATATYPE =>VALID_CERTIFICATE,
				     KEY=>MODIFIED ) . "\n";
print "   DELETED  => " . $db->getItem( DATATYPE =>VALID_CERTIFICATE,
				     KEY=>DELETED ) . "\n";
print "   ELEMENTS => " . $db->elements( DATATYPE => VALID_CERTIFICATE ) . "\n";
print "   NEXT ITEM SERIAL   => " . $db->getSerial(DATATYPE => VALID_CERTIFICATE).
									 "\n\n";

exit 0; 

